from __future__ import annotations

import torch
from torch import nn
from torchvision.models import resnet18  # type: ignore[import]

from rebasin import util
from rebasin.weight_matching.init_perms import PermutationInitializer
from rebasin.weight_matching.structs import AxisType

from .fixtures.models import MLP, ModuleWithWeirdWeightAndBiasNames


def test_init_permutations_weird_weight_bias_names() -> None:
    model_a = ModuleWithWeirdWeightAndBiasNames()
    model_b = ModuleWithWeirdWeightAndBiasNames()
    x = torch.randn(15)
    perm_init = PermutationInitializer(model_a, model_b, (x,))

    for perm in perm_init.permutations:
        names = [p.name for p in perm.parameters]
        if perm.parameters[0].name == "xyzweight":
            assert len(perm.parameters) == 1  # should not be associated with xyzbias

        if (
                perm.parameters[0].axis == 0
                and perm.parameters[0].name not in ("xyzweight", "xyzbias")
        ):
            assert len(perm.parameters) == 2  # every weight has a bias except xyz...
            if "weightabc" in names:
                assert "abcbias" in names
            if "defweightghi" in names:
                assert "defbiasghi" in names
            if "jklweight" in names:
                assert "biasjkl" in names


def test_init_permutations_multihead_attention() -> None:
    """Test the initialization of the permutations."""
    model_a = nn.MultiheadAttention(8, 8)
    model_b = nn.MultiheadAttention(8, 8)
    x = torch.randn(8, 8)
    perm_init = PermutationInitializer(model_a, model_b, (x, x, x))

    for permutation in perm_init.permutations:
        parameter_names = [p.name for p in permutation.parameters]
        for param_info in permutation.parameters:
            # Bias only on axis 0
            if "in_proj_weight" in parameter_names and param_info.axis == 0:
                assert "in_proj_bias" in parameter_names
                assert "out_proj.weight" not in parameter_names
            if "out_proj.weight" in parameter_names and param_info.axis == 0:
                assert "out_proj.bias" in parameter_names
                assert "in_proj_weight" not in parameter_names


def test_init_permutations_mlp() -> None:
    """Test the initialization of the permutations."""
    model_a = MLP(5)
    model_b = MLP(5)
    input_data = torch.randn(5)
    perm_init = PermutationInitializer(model_a, model_b, input_data)

    common_init_tests(model_a, model_b, perm_init)


def test_init_permutations_resnet() -> None:
    """Test the initialization of the permutations."""
    model_a = resnet18()
    model_b = resnet18()
    input_data = torch.randn(1, 3, 224, 224)
    perm_init = PermutationInitializer(model_a, model_b, input_data)

    common_init_tests(model_a, model_b, perm_init)


def common_init_tests(
        model_a: nn.Module, model_b: nn.Module, perm_init: PermutationInitializer
) -> None:
    modules = [
        m for m in model_b.modules()
    ]
    module_ids = [id(m) for m in modules]

    assert len(perm_init.permutations_init) > len(modules)  # several axes per module

    for permutation in perm_init.permutations:
        for param_info in permutation.parameters:
            assert param_info.module_id in perm_init.id_to_permutation_init
            assert param_info.module_id in module_ids
            assert util.contains_parameter(model_a.parameters(), param_info.param_a)
            assert util.contains_parameter(model_b.parameters(), param_info.param_b)

            assert permutation.perm_indices.shape[0] == \
                   param_info.param_a.shape[param_info.axis]

            assert permutation.perm_indices.shape[0] == \
                   param_info.param_b.shape[param_info.axis]

    # As several of the modules in the tested models are compatible,
    #   some permutations should be merged.
    assert len(perm_init.permutations) < len(perm_init.permutations_init)

    # Bias isn't turned off in either resnet18 or MLP
    #   -> axis_type = AxisType.NEITHER should be present.
    has_bias = any(
        "bias" in name
        for m in modules
        for name, _ in m.named_parameters()
    )

    if has_bias:
        assert any(
            param_info.axis_type == AxisType.NEITHER
            for permutation in perm_init.permutations
            for param_info in permutation.parameters
        )



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import Any

from torch import nn
from torch.utils.data import DataLoader


def recalculate_batch_norms(
        model: nn.Module,
        dataloader: DataLoader[Any],
        input_indices: int | Sequence[int],
        *forward_args: Any,
        **forward_kwargs: Any
) -> None:
    """
    Recalculate the BatchNorm statistics of a model.

    Use this after permuting the model, but only if the model contains
    a BatchNorm.

    Args:
        model:
            The model.
        dataloader:
            The DataLoader to use for recalculating the statistics.
            Should ideally be the training DataLoader.
        input_indices:
            Many DataLoaders return several inputs and labels.
            These can sometimes be used in unexpected ways.
            To make sure that the correct outputs of the dataloader are used
            as inputs to the model's forward pass, you can specify the indices
            at which the inputs are located, in the order that they should be
            passed to the model.
        *forward_args:
            Any additional positional arguments to pass to the model's forward  pass.
        **forward_kwargs:
            Any additional keyword arguments to pass to the model's forward pass.
    """
    # Don't waste compute on models without BatchNorm
    types = [type(m) for m in model.modules()]
    if (
            nn.BatchNorm1d not in types
            and nn.BatchNorm2d not in types
            and nn.BatchNorm3d not in types
    ):
        return

    training = model.training
    model.train()

    for batch in dataloader:
        if isinstance(batch, Sequence):
            inputs, _ = get_inputs_labels(batch, input_indices, 0)
        else:
            inputs = [batch]
        model(*inputs, *forward_args, **forward_kwargs)

    if not training:
        model.eval()


def get_inputs_labels(
        batch: Any,
        input_indices: int | Sequence[int] = 0,
        label_indices: int | Sequence[int] = 1
) -> tuple[list[Any], list[Any]]:
    """
    Get the inputs and outputs from a batch.

    Args:
        batch:
            The batch.
        input_indices:
            Many DataLoaders return several inputs and labels per batch.
            These can sometimes be used in unexpected ways.
            To make sure that the correct outputs of the dataloader are used
            as inputs to the model's forward pass, you can specify the indices
            at which the inputs are located, in the order that they should be
            passed to the model.
        label_indices:
            Like `input_indices`, but for the labels.

    Returns:
        The inputs and labels.
    """
    if isinstance(input_indices, int):
        input_indices = [input_indices]
    if isinstance(label_indices, int):
        label_indices = [label_indices]

    inputs = [batch[i] for i in input_indices]
    labels = [batch[i] for i in label_indices]
    return inputs, labels


def contains_parameter(
        parameters: Sequence[nn.Parameter] | Iterator[nn.Parameter],
        parameter: nn.Parameter
) -> bool:
    """
    Check if a sequence of parameters contains a parameter.

    This cannot be done via the normal `in` operator, because
    `nn.Parameter`'s `__eq__` does not work for parameters of different shapes.

    Args:
        parameters:
            The sequence of parameters.
        parameter:
            The parameter.

    Returns:
        Whether the sequence contains the parameter.
    """
    return any(param is parameter for param in parameters)



from __future__ import annotations

import copy
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import torch
from torch import nn
from torch.utils.data import DataLoader

from .util import get_inputs_labels, recalculate_batch_norms


class Interpolation:
    def __init__(
            self,
            models: Sequence[torch.nn.Module],
            loss_fn: Any,
            train_dataloader: DataLoader[Any],
            val_dataloader: DataLoader[Any],
            input_indices: Sequence[int] | int = 0,
            output_indices: Sequence[int] | int = 1,
            savedir: Path | str | None = None,
            save_all: bool = False,
    ) -> None:
        self._sanity_checks(
            models,
            loss_fn,
            train_dataloader,
            val_dataloader,
            input_indices,
            output_indices,
            savedir,
            save_all

        )
        if savedir is not None and isinstance(savedir, str):
            savedir = Path(savedir)

        self.models = models
        self.loss_fn = loss_fn
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.input_indices = input_indices
        self.output_indices = output_indices
        self.savedir = savedir
        self.save_all = save_all

        self.losses_original = [self.evaluate_model(model) for model in self.models]
        self.losses_interpolated: list[float] = []

        min_idx = int(torch.argmin(torch.tensor(self.losses_original)))
        self.best_loss = self.losses_original[min_idx]
        self.best_model = self.models[min_idx]
        self.best_model_name = f"model_{min_idx}.pt"

    def interpolate(
            self,
            steps: int = 20,
            savedir: str | Path | None = None,
            save_all: bool = False
    ) -> None:
        """Interpolate between the models and save the best one or all."""

    def evaluate_model(self, model: nn.Module) -> float:
        model.eval()
        loss = 0.0
        for batch in self.val_dataloader:
            inputs, labels = get_inputs_labels(
                batch, self.input_indices, self.output_indices
            )
            y_pred = model(*inputs)
            loss += float(self.loss_fn(y_pred, *labels))
        return loss / len(self.val_dataloader)

    @staticmethod
    def _sanity_checks(
            models: Sequence[torch.nn.Module],
            loss_fn: Any,
            train_dataloader: DataLoader[Any],
            val_dataloader: DataLoader[Any],
            input_indices: Sequence[int] | int,
            output_indices: Sequence[int] | int,
            savedir: Path | str | None,
            save_all: bool = False,
    ) -> None:
        assert isinstance(models, Sequence), "Models must be a sequence"
        assert all(isinstance(model, nn.Module) for model in models), \
            "All models must be a subclass of nn.Module"

        assert callable(loss_fn), "Loss function must be callable"

        assert isinstance(train_dataloader, DataLoader)
        assert isinstance(val_dataloader, DataLoader)

        assert isinstance(input_indices, (int, Sequence))
        assert isinstance(output_indices, (int, Sequence))

        if isinstance(input_indices, Sequence):
            assert all(isinstance(i, int) for i in input_indices)
        if isinstance(output_indices, Sequence):
            assert all(isinstance(i, int) for i in output_indices)

        assert isinstance(savedir, (Path, str, type(None)))
        assert isinstance(save_all, bool)

        assert len(models) > 1, "Need at least two models to interpolate between"

        if save_all:
            assert savedir is not None


class LerpSimple(Interpolation):
    """
    Linear interpolation between two models.
    """

    def interpolate(
            self,
            steps: int = 20,
            savedir: str | Path | None = None,
            save_all: bool | None = None
    ) -> None:
        r"""
        Interpolate between the models and save the best one or all
        in :code:`self.best_model`.

        For two models and `s` steps, the interpolation looks like this
        (using :math:`m_a` to denote model number :math:`a`,
        and :math:`i_{ab,c}` to denote interpolation step `c`
        between models number `a` and `b`):

        |

        :math:`m_1 - i_{12,1} - ... - i_{12,s} - m_2`

        |

        For `n` models, the same is repeated between every other pair of models,
        meaning:

        |

        :math:`m_1 - i_{12,1} - ... - i_{12,s} - m_2
        - i_{23,1} - ... - i_{23,s} - m_3 - ... - m_n`

        |

        This means that the total number of interpolations is :math:`s \cdot (n - 1)`.

        |

        Args:
            steps:
                The number of steps to take between each pair of models.

            savedir:
                The directory to save the models in. Overwrites the :code:`savedir`
                argument passed to the constructor if not None.
                If both are :code:`None`, the models are not saved.

                If :code:`not None`, the following naming schema is used for the files:

                - :code:`"model_{model_num}.pt"`
                    if the model is one of the original models.
                    :code:`model_num` is the index of the model
                    in the :code:`models` argument.
                    The original models are not saved, except the best model.

                - :code:`"interp_models_{m_i}_{m_i+1}_perc_{percentage}.pt"`
                  if the model is interpolated.
                  :math:`m_i` and :math:`m_{i+1}` are the indices of the models
                  between which the interpolation is done.
                  `percentage` is the percentage of the way through the interpolation.
                  If it is small, the model is closer to :math:`m_i`,
                  and if it is large, the model is closer to :math:`m_{i+1}`.

            save_all:
                Whether to save all interpolated models, or just the best one.
                The best model is saved either way if a `savedir` is given.
                Overwrites the `save_all` argument passed to the constructor
                if not None.
        """
        # SANITY CHECKS AND DEFAULT SETTINGS
        assert isinstance(savedir, (Path, str, type(None)))

        if savedir is None:
            savedir = self.savedir
        elif isinstance(savedir, str):
            savedir = Path(savedir)

        if save_all is None:
            save_all = self.save_all

        assert isinstance(save_all, bool)

        # INTERPOLATION
        for step in range(steps):
            # Interpolate between the two models
            # (step + 1) so that it never starts at zero percent.
            # (steps + 2) so that it never ends at 100 percent.
            # This is because the start and end models already exist
            percentage = (step + 1) / (steps + 2)
            self._interpolate_step(
                percentage=percentage, savedir=savedir, save_all=save_all
            )

        # SAVE THE BEST MODEL
        # If all interpolated models are saved and the best model is interpolated,
        #   it is unnecessary to save it again.
        # Conversely, if not all interpolated models are saved, the best model
        #   must be saved.
        # Alternatively, save_all only leads to interpolated models being saved,
        #   so the best model must be saved again if it is not interpolated.
        save_best_model = (not save_all) or (self.best_model in self.models)

        if save_best_model and (savedir is not None):
            torch.save(
                self.best_model.state_dict(),
                savedir.joinpath(self.best_model_name)
            )

    def _interpolate_step(
            self, percentage: float, savedir: Path | None, save_all: bool
    ) -> None:
        # Interpolate between all models
        for model_num, (model1, model2) in enumerate(
                zip(self.models[:-1], self.models[1:], strict=True)
        ):
            model_interp = copy.deepcopy(model1)
            for param1, param2, param_interp in zip(
                    model1.parameters(),
                    model2.parameters(),
                    model_interp.parameters(),
                    strict=True
            ):
                param_interp.data = torch.lerp(param1, param2, percentage)

            # Recalculate BatchNorm statistics
            recalculate_batch_norms(
                model_interp, self.train_dataloader, self.input_indices
            )

            # Evaluate
            loss = self.evaluate_model(model_interp)
            self.losses_interpolated.append(loss)

            # Save
            filename = f"interp_models_{model_num}_{model_num+1}_perc_{percentage}.pt"

            if loss < self.best_loss:
                self.best_loss = loss
                self.best_model = model_interp
                self.best_model_name = filename  # for later saving if not save_all

            if savedir is not None and save_all:
                torch.save(model_interp.state_dict(), savedir.joinpath(filename))



from __future__ import annotations

import copy
from typing import Any

import torch
from torch import nn
from torchvision.models import resnet18  # type: ignore[import]

from rebasin import PermutationCoordinateDescent
from rebasin.weight_matching.permutation_coordinate_descent import calculate_progress
from tests.fixtures.models import MLP, ModuleWithWeirdWeightAndBiasNames


def test_calculate_progress() -> None:
    """Test the calculation of the progress."""
    cost_mat = torch.tensor([[0.0, 1.0, 2.0], [1.0, 0.0, 3.0], [2.0, 3.0, 0.0]])
    perm_old = torch.tensor([0, 1, 2])
    perm_new = torch.tensor([2, 0, 1])

    progress = calculate_progress(cost_mat, perm_old, perm_new)

    assert progress

    perm_old = torch.tensor([2, 0, 1])
    perm_new = torch.tensor([0, 1, 2])

    progress = calculate_progress(cost_mat, perm_old, perm_new)

    assert not progress


class TestPermutationCoordinateDescent:
    def test_resnet18(self) -> None:
        # In a run of 500 tests, this test succeeded every time.
        model_a = resnet18()
        model_b = resnet18()
        self.common_tests(model_a, model_b, torch.randn(1, 3, 224, 224))

    def test_mlp(self) -> None:
        # Wide layers, or else some permutations will be the identity.
        # If every permutation is the identity, the test will fail.
        # Increasing layer size is better than increasing layer count.
        # At weights of size (25, 25), the test succeeds 500 out of 500 attempts.
        # For (5, 5) weights, the test fails fairly often.
        model_a = MLP(25, num_layers=5)
        model_b = MLP(25, num_layers=5)
        self.common_tests(model_a, model_b, torch.randn(25))

    def test_multihead_attention(self) -> None:
        embed_dim = num_heads = 8

        model_a = nn.MultiheadAttention(embed_dim, num_heads)
        model_b = nn.MultiheadAttention(embed_dim, num_heads)

        query = torch.randn(embed_dim, num_heads, requires_grad=True)
        key = torch.randn(embed_dim, num_heads, requires_grad=True)
        value = torch.randn(embed_dim, num_heads, requires_grad=True)

        self.common_tests(model_a, model_b, (query, key, value))

    def test_module_with_weird_weight_and_bias_names(self) -> None:
        model_a = ModuleWithWeirdWeightAndBiasNames()
        model_b = ModuleWithWeirdWeightAndBiasNames()

        self.common_tests(model_a, model_b, torch.randn(15))

    def common_tests(
            self,
            model_a: nn.Module,
            model_b: nn.Module,
            input_data: Any,
    ) -> None:
        pcd = PermutationCoordinateDescent(model_a, model_b, input_data)
        pcd.calculate_permutations()
        model_b_old = copy.deepcopy(model_b)  # for comparison

        # Check that there are some permutations different from the identity
        is_arange_list: list[bool] = []
        for permutation in pcd.permutations:
            ci, n = permutation.perm_indices, len(permutation.perm_indices)
            is_arange_list.append(bool(torch.all(torch.eq(ci, torch.arange(n))).item()))

        assert not all(is_arange_list)

        # Check that:
        # - model_b was actually changed - I don't have to manually assing results
        # - model_b is closer to model_a than it was before the optimization
        pcd.apply_permutations()
        assert (
                self.model_similarity(model_a, model_b)
                > self.model_similarity(model_a, model_b_old)
        )

    @staticmethod
    def model_similarity(model_a: nn.Module, model_b: nn.Module) -> float:
        """Calculate the distance between two models."""
        total_dist = 0.0

        for parameter_a, parameter_b in zip(
                model_a.parameters(), model_b.parameters(), strict=True
        ):
            p_a, p_b = parameter_a.reshape(-1), parameter_b.reshape(-1)
            total_dist += float(p_a @ p_b)
        return abs(total_dist)



from __future__ import annotations

import pytest
import torch
from torch import nn
from torch.utils.data import DataLoader

from rebasin import interpolation as interp
from tests.fixtures.datasets import RandomDataset
from tests.fixtures.models import MLP


def loss_fn(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return ((x - y) ** 2).sum()  # type: ignore[no-any-return]


def test_loss_fn() -> None:
    """Test that loss_fn is working as expected."""
    l1 = loss_fn(torch.tensor([1.0]), torch.tensor([2.0]))
    l2 = loss_fn(torch.tensor([1.0]), torch.tensor([1.0]))
    assert l1 > l2


class TestInterpolation:
    """Test sanity checks of Interpolation class."""
    @property
    def models(self) -> list[nn.Module]:
        return [MLP(5), MLP(5)]

    @property
    def train_dataloader(
            self
    ) -> torch.utils.data.DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return DataLoader(RandomDataset(shape=(5,), length=10))

    @property
    def val_dataloader(
            self
    ) -> torch.utils.data.DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return DataLoader(RandomDataset(shape=(5,), length=4))

    def test_sanity_checks_model(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=[],
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=MLP(5),  # type: ignore[arg-type]
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                input_indices=1
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=[*self.models, "not a model"],  # type: ignore[list-item]
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                input_indices=1
            )

    def test_sanity_checks_loss_fn(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=0,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader
            )

    def test_sanity_checks_dataloaders(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=0,  # type: ignore[arg-type]
                val_dataloader=self.val_dataloader
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=0  # type: ignore[arg-type]
            )

    def test_sanity_checks_input_indices(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                input_indices=0.5  # type: ignore[arg-type]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                input_indices=[0.5]  # type: ignore[list-item]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                input_indices=["not", "an", "int"]  # type: ignore[list-item]
            )

    def test_sanity_checks_output_indices(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                output_indices=0.5  # type: ignore[arg-type]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                output_indices=[0.5]  # type: ignore[list-item]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                output_indices=["not", "an", "int"]  # type: ignore[list-item]
            )

    def test_sanity_checks_save_vars(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                savedir=0  # type: ignore[arg-type]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                save_all="not a bool"  # type: ignore[arg-type]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                loss_fn=loss_fn,
                train_dataloader=self.train_dataloader,
                val_dataloader=self.val_dataloader,
                save_all=True,
                savedir=None
            )


class TestLerpSimple:
    """Test the LerpSimple class."""

    @property
    def mlps(self) -> list[nn.Module]:
        return [MLP(5), MLP(5)]

    @property
    def train_dl_mlp(
            self
    ) -> torch.utils.data.DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return DataLoader(RandomDataset(shape=(5,), length=10))

    @property
    def val_dl_mlp(
            self
    ) -> torch.utils.data.DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return DataLoader(RandomDataset(shape=(5,), length=2))

    def test_settings_are_plausible(self) -> None:
        model_a, model_b = self.mlps

        # Must be different models
        for p1, p2 in zip(model_a.parameters(), model_b.parameters(), strict=True):
            assert not torch.allclose(p1, p2)

        # Must produce different outputs
        x = torch.rand(5)
        assert not torch.allclose(model_a(x), model_b(x))

        # Must have different losses
        val_dl = self.val_dl_mlp
        x, y = next(iter(val_dl))
        loss_a = loss_fn(model_a(x), y)
        loss_b = loss_fn(model_b(x), y)

        assert loss_a != loss_b

    def test_mlp(self) -> None:
        lerp = interp.LerpSimple(
            models=self.mlps,
            train_dataloader=self.train_dl_mlp,
            val_dataloader=self.val_dl_mlp,
            loss_fn=loss_fn
        )
        lerp.interpolate(steps=10)

        assert len(lerp.losses_interpolated) == 10
        assert len(lerp.losses_original) == 2

        # With random data, there should be no duplicate losses.
        assert len(set(lerp.losses_original)) == 2
        assert len(set(lerp.losses_interpolated)) == 10
